#!/usr/bin/env python3.6
#
# Copyright 2018 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"secrets tests"

import os
import unittest
import re
import tempfile
import gnupg
from kapitan.secrets import secret_token_attributes, SECRET_TOKEN_TAG_PATTERN
from kapitan.secrets import secret_gpg_write, secret_gpg_reveal_raw
from kapitan.secrets import secret_gpg_update_recipients, secret_gpg_raw_read_fingerprints

GPG_HOME = tempfile.mkdtemp()
GPG_OBJ = gnupg.GPG(gnupghome=GPG_HOME)
SECRETS_HOME = tempfile.mkdtemp()
KEY = GPG_OBJ.gen_key(GPG_OBJ.gen_key_input(key_type="RSA",
                                            key_length=2048,
                                            passphrase="testphrase"))
KEY2 = GPG_OBJ.gen_key(GPG_OBJ.gen_key_input(key_type="RSA",
                                            key_length=2048,
                                            passphrase="testphrase"))


class SecretsTest(unittest.TestCase):
    "Test secrets"
    def test_secret_token_attributes(self):
        "grab attributes and compare to values"
        token_tag = '?{gpg:secret/sauce}'
        _token_tag, token, func = re.match(SECRET_TOKEN_TAG_PATTERN,
                                           token_tag).groups()
        self.assertEqual(_token_tag, token_tag)
        backend, token_path = secret_token_attributes(token)
        self.assertEqual((backend, token_path), ('gpg', 'secret/sauce'))

    def test_gpg_secret_write_reveal(self):
        "write secret, confirm secret file exists, reveal and compare content"
        token = 'secret/sauce'
        secret_gpg_write(GPG_OBJ, SECRETS_HOME, token, "super secret value",
                         False, [{'fingerprint': KEY.fingerprint}],
                         passphrase="testphrase")
        self.assertTrue(os.path.isfile(os.path.join(SECRETS_HOME, token)))

        file_with_secret_tags = tempfile.mktemp()
        file_revealed = tempfile.mktemp()
        with open(file_with_secret_tags, 'w') as fp:
            fp.write('I am a file with a ?{gpg:secret/sauce:deadbeef}')
        with open(file_revealed, 'w') as fp:
            secret_gpg_reveal_raw(GPG_OBJ, SECRETS_HOME, file_with_secret_tags,
                              verify=False, output=fp, passphrase="testphrase")
        with open(file_revealed) as fp:
            self.assertEqual("I am a file with a super secret value", fp.read())

    def test_gpg_secret_base64_write_reveal(self):
        """
        write secret for base64 encoded content, confirm secret file exists,
        reveal and compare content
        """
        token = 'secret/sauce'
        secret_gpg_write(GPG_OBJ, SECRETS_HOME, token, "super secret value",
                         True, [{'fingerprint': KEY.fingerprint}],
                         passphrase="testphrase")
        self.assertTrue(os.path.isfile(os.path.join(SECRETS_HOME, token)))

        file_with_secret_tags = tempfile.mktemp()
        file_revealed = tempfile.mktemp()
        with open(file_with_secret_tags, 'w') as fp:
            fp.write('I am a file with a ?{gpg:secret/sauce:deadbeef}')
        with open(file_revealed, 'w') as fp:
            secret_gpg_reveal_raw(GPG_OBJ, SECRETS_HOME, file_with_secret_tags,
                              verify=False, output=fp, passphrase="testphrase")
        with open(file_revealed) as fp:
            self.assertEqual("I am a file with a c3VwZXIgc2VjcmV0IHZhbHVl", fp.read())

    def test_gpg_secret_update_recipients(self):
        """
        update existing secret with another recipient, confirm content is the same
        """
        token = 'secret/sauce'
        secret_gpg_write(GPG_OBJ, SECRETS_HOME, token, "super secret value",
                         True, [{'fingerprint': KEY.fingerprint}],
                         passphrase="testphrase")
        self.assertTrue(os.path.isfile(os.path.join(SECRETS_HOME, token)))
        self.assertTrue(len(secret_gpg_raw_read_fingerprints(SECRETS_HOME, token)), 1)

        file_with_secret_tags = tempfile.mktemp()
        file_revealed = tempfile.mktemp()
        with open(file_with_secret_tags, 'w') as fp:
            fp.write('I am a file with a ?{gpg:secret/sauce:deadbeef}')

        new_recipients = [{'fingerprint': KEY.fingerprint},
                          {'fingerprint': KEY2.fingerprint}]
        secret_gpg_update_recipients(GPG_OBJ, SECRETS_HOME, token, new_recipients,
                                     passphrase="testphrase")
        self.assertTrue(len(secret_gpg_raw_read_fingerprints(SECRETS_HOME, token)), 2)
        with open(file_revealed, 'w') as fp:
            secret_gpg_reveal_raw(GPG_OBJ, SECRETS_HOME, file_with_secret_tags,
                              verify=False, output=fp, passphrase="testphrase")
        with open(file_revealed) as fp:
            self.assertEqual("I am a file with a c3VwZXIgc2VjcmV0IHZhbHVl", fp.read())



#!/usr/bin/env python3.6
#
# Copyright 2018 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"kapitan resources"

import errno
from functools import partial
import ujson as json
import logging
import os
import io
import reclass
import reclass.core
from reclass.errors import ReclassException, NotFoundError
import yaml

from kapitan.utils import render_jinja2_file, sha256_string
from kapitan import __file__ as kapitan_install_path
from kapitan.errors import CompileError, InventoryError
import kapitan.cached as cached

logger = logging.getLogger(__name__)

try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader


def resource_callbacks(search_path):
    """
    Returns a dict with all the functions to be used
    on the native_callbacks keyword on jsonnet
    search_path can be used by the native functions to access files
    """

    return {"jinja2_render_file": (("name", "ctx"),
                                   partial(jinja2_render_file, search_path)),
            "inventory": (("target", "inv_path"),
                          partial(inventory, search_path)),
            "file_read": (("name",),
                          partial(read_file, search_path)),
            "yaml_dump": (("obj",), yaml_dump),
            "sha256_string": (("obj",), sha256_string),
           }


def yaml_dump(obj):
    "Dumps jsonnet obj as yaml"
    _obj = json.loads(obj)
    return yaml.safe_dump(_obj, default_flow_style=False)


def jinja2_render_file(search_path, name, ctx):
    """
    Render jinja2 file name with context ctx.
    search_path is used to find the file name
    as there is a limitation with jsonnet's native_callback approach:
    one can't access the current directory being evaluated
    """
    ctx = json.loads(ctx)
    _full_path = os.path.join(search_path, name)
    logger.debug("jinja2_render_file trying file %s", _full_path)
    try:
        if os.path.exists(_full_path):
            logger.debug("jinja2_render_file found file at %s", _full_path)
            return render_jinja2_file(_full_path, ctx)
        else:
            raise IOError("Could not find file %s" % name)
    except Exception as e:
        logger.error("Jsonnet jinja2 failed to render %s: %s", _full_path, str(e))
        raise CompileError(e)


def read_file(search_path, name):
    "return content of file in name"
    full_path = os.path.join(search_path, name)
    logger.debug("read_file trying file %s", full_path)
    if os.path.exists(full_path):
        logger.debug("read_file found file at %s", full_path)
        with io.open(full_path, newline='') as f:
            return f.read()
    raise IOError("Could not find file %s" % name)


def search_imports(cwd, import_str, search_path):
    """
    This is only to be used as a callback for the jsonnet API!
    - cwd is the import context $CWD,
    - import_str is the "bla.jsonnet" in 'import "bla.jsonnet"'
    - search_path is the location where to look for import_str if not in cwd
    The only supported parameters are cwd and import_str, so search_path
    needs to be closured.
    """
    basename = os.path.basename(import_str)
    full_import_path = os.path.join(cwd, import_str)

    # if import_str not found, search in search_path
    if not os.path.exists(full_import_path):
        install_path = os.path.dirname(kapitan_install_path)
        for path in (install_path, search_path):
            _full_import_path = os.path.join(path, import_str)
            # if found, set as full_import_path
            if os.path.exists(_full_import_path):
                full_import_path = _full_import_path
                logger.debug("import_str: %s found in search_path: %s",
                             import_str, path)
                break

    # if the above search did not find anything, let jsonnet error
    # with a non existent import
    normalised_path = os.path.normpath(full_import_path)

    logger.debug("cwd:%s import_str:%s basename:%s -> norm:%s",
                 cwd, import_str, basename, normalised_path)

    normalised_path_content = ""
    with open(normalised_path) as f:
        normalised_path_content = f.read()

    return normalised_path, normalised_path_content


def inventory(search_path, target, inventory_path="inventory/"):
    """
    Reads inventory (set by inventory_path) in search_path.
    set nodes_uri to change reclass nodes_uri the default value
    set target to None to return all target in the inventory
    Returns a dictionary with the inventory for target
    """
    full_inv_path = os.path.join(search_path, inventory_path)

    if target is None:
        return inventory_reclass(full_inv_path)["nodes"]

    return inventory_reclass(full_inv_path)["nodes"][target]


def inventory_reclass(inventory_path):
    """
    Runs a reclass inventory in inventory_path
    (same output as running ./reclass.py -b streams/ --inventory)
    Will attempt to read reclass config from 'reclass-config.yml' otherwise
    it will failback to the default config.
    Returns a reclass style dictionary
    """

    if not cached.inv:
        reclass_config = {'storage_type': 'yaml_fs',
                          'inventory_base_uri': inventory_path,
                          'nodes_uri': os.path.join(inventory_path, 'targets'),
                          'classes_uri': os.path.join(inventory_path, 'classes')
                         }

        try:
            cfg_file = os.path.join(inventory_path, 'reclass-config.yml')
            with open(cfg_file) as reclass_cfg:
                reclass_config = yaml.load(reclass_cfg, Loader=YamlLoader)
                # normalise relative nodes_uri and classes_uri paths
                for uri in ('nodes_uri', 'classes_uri'):
                    uri_val = reclass_config.get(uri)
                    uri_path = os.path.join(inventory_path, uri_val)
                    normalised_path = os.path.normpath(uri_path)
                    reclass_config.update({uri: normalised_path})
                logger.debug("Using reclass inventory config at: %s", cfg_file)
        except IOError as ex:
            # If file does not exist, ignore
            if ex.errno == errno.ENOENT:
                logger.debug("Using reclass inventory config defaults")

        try:
            storage = reclass.get_storage(reclass_config['storage_type'], reclass_config['nodes_uri'],
                                          reclass_config['classes_uri'])
            class_mappings = reclass_config.get('class_mappings')  # this defaults to None (disabled)
            _reclass = reclass.core.Core(storage, class_mappings, reclass.settings.Settings())

            cached.inv = _reclass.inventory()
            return cached.inv
        except ReclassException as e:
            if isinstance(e, NotFoundError):
                logger.error("Inventory reclass error: inventory not found")
            else:
                logger.error("Inventory reclass error: %s", e.message)
            raise InventoryError(e.message)
    else:
        return cached.inv



#!/usr/bin/env python3.6
#
# Copyright 2018 The Kapitan Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

"random utils"

from functools import reduce
from hashlib import sha256
import logging
import os
import sys
import stat
import collections
import jinja2
import _jsonnet as jsonnet
import yaml
from distutils.version import StrictVersion

from kapitan.version import VERSION
from kapitan.errors import CompileError

logger = logging.getLogger(__name__)

try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader


class termcolor:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def normalise_join_path(dirname, path):
    "Join dirname with path and return in normalised form"
    logger.debug(os.path.normpath(os.path.join(dirname, path)))
    return os.path.normpath(os.path.join(dirname, path))


def render_jinja2_template(content, context):
    "Render jinja2 content with context"
    return jinja2.Template(content, undefined=jinja2.StrictUndefined).render(context)


def sha256_string(string):
    "Returns sha256 hex digest for string"
    return sha256(string.encode("UTF-8")).hexdigest()


def jinja2_yaml_filter(obj):
    "Returns yaml for object"
    return yaml.safe_dump(obj, default_flow_style=False)


def render_jinja2_file(name, context):
    "Render jinja2 file name with context"
    path, filename = os.path.split(name)
    env = jinja2.Environment(
        undefined=jinja2.StrictUndefined,
        loader=jinja2.FileSystemLoader(path or './'),
        trim_blocks=True,
        lstrip_blocks=True,
        extensions=['jinja2.ext.do'],
    )
    env.filters['sha256'] = sha256_string
    env.filters['yaml'] = jinja2_yaml_filter
    return env.get_template(filename).render(context)


def render_jinja2(path, context):
    """
    Render files in path with context
    Returns a dict where the is key is the filename (with subpath)
    and value is a dict with content and mode
    Empty paths will not be rendered
    Path can be a single file or directory
    Ignores hidden files (.filename)
    """
    rendered = {}
    walk_root_files = []
    if os.path.isfile(path):
        dirname = os.path.dirname(path)
        basename = os.path.basename(path)
        walk_root_files = [(dirname, None, [basename])]
    else:
        walk_root_files = os.walk(path)

    for root, _, files in walk_root_files:
        for f in files:
            if f.startswith('.'):
                logger.debug('render_jinja2: ignoring file %s', f)
                continue
            render_path = os.path.join(root, f)
            logger.debug("render_jinja2 rendering %s", render_path)
            # get subpath and filename, strip any leading/trailing /
            name = render_path[len(os.path.commonprefix([root, path])):].strip('/')
            try:
                rendered[name] = {
                    "content": render_jinja2_file(render_path, context),
                    "mode": file_mode(render_path)
                }
            except Exception as e:
                logger.error("Jinja2 error: failed to render %s: %s", render_path, str(e))
                raise CompileError(e)
    return rendered


def file_mode(name):
    "Returns mode for file name"
    st = os.stat(name)
    return stat.S_IMODE(st.st_mode)


def jsonnet_file(file_path, **kwargs):
    """
    Evaluate file_path jsonnet file.
    kwargs are documented in http://jsonnet.org/implementation/bindings.html
    """
    try:
        return jsonnet.evaluate_file(file_path, **kwargs)
    except Exception as e:
        logger.error("Jsonnet error: failed to compile %s:\n %s", file_path, str(e))
        raise CompileError(e)


def prune_empty(d):
    '''Remove empty lists and empty dictionaries from x.
    (similar to jsonnet std.prune but faster)
    '''
    if not isinstance(d, (dict, list)):
        return d

    if isinstance(d, list):
        if len(d) > 0:
            return [v for v in (prune_empty(v) for v in d) if v is not None]

    if isinstance(d, dict):
        if len(d) > 0:
            return {k: v for k, v in ((k, prune_empty(v)) for k, v in d.items()) if v is not None}


class PrettyDumper(yaml.SafeDumper):
    '''
    Increases indent of nested lists.
    By default, they are indendented at the same level as the key on the previous line
    More info on https://stackoverflow.com/questions/25108581/python-yaml-dump-bad-indentation
    '''
    def increase_indent(self, flow=False, indentless=False):
        return super(PrettyDumper, self).increase_indent(flow, False)


def flatten_dict(d, parent_key='', sep='.'):
    '''
    Flatten nested elements in a dictionary
    '''
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def deep_get(dictionary, keys, previousKey=None):
    '''
    Search recursively for 'keys' in 'dictionary' and return value, otherwise return None
    '''
    value = None
    if len(keys) > 0:
        value = dictionary.get(keys[0], None) if isinstance(dictionary, dict) else None

        if value:
            # If we are at the last key and we have a value, we are done
            if len(keys) == 1:
                return value

            # If we have more variables in the search chain and we don't have a dict, return not found
            if not isinstance(value, dict):
                return None

            # Recurse with next keys in the chain on the dict
            return deep_get(value, keys[1:], previousKey=keys[0])
        else:
            if isinstance(dictionary, dict):
                # If we find nothing, check for globbing, loop and match with dict keys
                if '*' in keys[0]:
                    key_lower = keys[0].replace('*', '').lower()
                    for dict_key in dictionary.keys():
                        if key_lower in dict_key.lower():
                            # If we're at the last key in the chain, return matched value
                            if len(keys) == 1:
                                return dictionary[dict_key]

                            # If we have more variables in the chain, continue recursion
                            return deep_get(dictionary[dict_key], keys[1:], previousKey=keys[0])

                if not previousKey:
                    # No previous keys in chain and no globbing, move down the dictionary and recurse
                    for v in dictionary.values():
                        if isinstance(v, dict):
                            item = None
                            if len(keys) > 1:
                                item = deep_get(v, keys, previousKey=keys[0])
                            else:
                                item = deep_get(v, keys)

                            if item:
                                return item

    return value


def searchvar(flat_var, inventory_path):
    '''
    show all inventory files where a given reclass variable is declared
    '''

    output = []
    maxlenght = 0
    keys = flat_var.split(".")
    for root, _, files in os.walk(inventory_path):
        for file in files:
            if file.endswith(".yml") or file.endswith(".yaml"):
                filename = os.path.join(root, file)
                with open(filename, 'r') as fd:
                    data = yaml.load(fd, Loader=YamlLoader)
                    value = deep_get(data, keys)
                    if value is not None:
                        output.append((filename, value))
                        if len(filename) > maxlenght:
                            maxlenght = len(filename)
    for i in output:
        print('{0!s:{l}} {1!s}'.format(*i, l=maxlenght + 2))


def get_directory_hash(directory):
    '''
    Compute a sha256 hash for the file contents of a directory
    '''
    if not os.path.exists(directory):
        logger.error("utils.get_directory_hash failed, %s dir doesn't exist", directory)
        return -1

    try:
        hash = sha256()
        for root, _, files in os.walk(directory):
            for names in files:
                file_path = os.path.join(root, names)
                try:
                    with open(file_path, 'r') as f:
                        hash.update(sha256(f.read().encode("UTF-8")).hexdigest().encode("UTF-8"))
                except Exception as e:
                    logger.error("utils.get_directory_hash failed to open %s: %s", file_path, str(e))
                    raise

    except Exception as e:
        logger.error("utils.get_directory_hash failed: %s", str(e))
        raise

    return hash.hexdigest()


def check_version():
    '''
    Checks that the last version of kapitan used is at least smaller or equal to current version.
    If the last version of kapitan used is bigger, it will give instructions on how to upgrade and exit(1).
    '''
    if os.path.exists('.kapitan'):
        with open('.kapitan', 'r') as f:
            dot_kapitan = yaml.safe_load(f)
            # If 'saved version is bigger than current version'
            if dot_kapitan['version'] and StrictVersion(dot_kapitan['version']) > StrictVersion(VERSION):
                print('{}Current version: {}'.format(termcolor.WARNING, VERSION))
                print('Last used version (in .kapitan): {}{}\n'.format(dot_kapitan["version"], termcolor.ENDC))
                print('Please upgrade kapitan to at least "{}" in order to keep results consistent:\n'.format(dot_kapitan["version"]))
                print('Docker: docker pull deepmind/kapitan')
                print('Pip (user): pip3 install --user --upgrade kapitan\n')
                print('Check https://github.com/deepmind/kapitan#quickstart for more info.\n')
                print('If you know what you\'re doing, you can skip this check by adding \'--ignore-version-check\'.')
                sys.exit(1)


def save_version():
    '''
    Saves the current kapitan version to a local .kapitan file
    '''
    with open('.kapitan', 'w') as f:
        dot_kapitan = {'version': VERSION}
        yaml.safe_dump(dot_kapitan, stream=f, default_flow_style=False)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#

from reclass.settings import Settings
from .item import Item

class CompItem(Item):

    def __init__(self, items, settings):
        self.type = Item.COMPOSITE
        self._items = items
        self._settings = settings
        self._refs = []
        self._allRefs = False
        self.assembleRefs()

    def assembleRefs(self, context={}):
        self._refs = []
        self._allRefs = True
        for item in self._items:
            if item.has_references():
                item.assembleRefs(context)
                self._refs.extend(item.get_references())
                if item.allRefs() is False:
                    self._allRefs = False

    def contents(self):
        return self._items

    def allRefs(self):
        return self._allRefs

    def has_references(self):
        return len(self._refs) > 0

    def get_references(self):
        return self._refs

    def merge_over(self, item):
        if item.type == Item.SCALAR or item.type == Item.COMPOSITE:
            return self
        elif item.type == Item.LIST:
            if self._settings.allow_scalar_over_list or (self._settings.allow_none_override and self._value in [None, 'none', 'None']):
                return self
            else:
                raise TypeError('allow scalar over list = False: cannot merge %s over %s' % (repr(self), repr(item)))
        elif item.type == Item.DICTIONARY:
            if self._settings.allow_scalar_over_dict or (self._settings.allow_none_override and self._value in [None, 'none', 'None']):
                return self
            else:
                raise TypeError('allow scalar over dict = False: cannot merge %s over %s' % (repr(self), repr(item)))
        raise TypeError('Cannot merge %s over %s' % (repr(self), repr(item)))

    def render(self, context, inventory):
        # Preserve type if only one item
        if len(self._items) == 1:
            return self._items[0].render(context, inventory)
        # Multiple items
        strings = [ str(i.render(context, inventory)) for i in self._items ]
        return "".join(strings)

    def __repr__(self):
        return 'CompItem(%r)' % self._items

    def __str__(self):
        return ''.join([ str(i) for i in self._items ])



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#

from reclass.settings import Settings
from .item import Item

class ScaItem(Item):

    def __init__(self, value, settings):
        self.type = Item.SCALAR
        self._value = value
        self._settings = settings

    def contents(self):
        return self._value

    def merge_over(self, item):
        if item.type == Item.SCALAR or item.type == Item.COMPOSITE:
            return self
        elif item.type == Item.LIST:
            if self._settings.allow_scalar_over_list or (self._settings.allow_none_override and self._value is None):
                return self
            else:
                raise TypeError('allow scalar over list = False: cannot merge %s over %s' % (repr(self), repr(item)))
        elif item.type == Item.DICTIONARY:
            if self._settings.allow_scalar_over_dict or (self._settings.allow_none_override and self._value is None):
                return self
            else:
                raise TypeError('allow scalar over dict = False: cannot merge %s over %s' % (repr(self), repr(item)))
        raise TypeError('Cannot merge %s over %s' % (repr(self), repr(item)))

    def render(self, context, inventory):
        return self._value

    def __repr__(self):
        return 'ScaItem({0!r})'.format(self._value)

    def __str__(self):
        return str(self._value)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#

import copy
import pyparsing as pp

from six import iteritems

from .item import Item
from reclass.settings import Settings
from reclass.utils.dictpath import DictPath
from reclass.errors import ExpressionError, ParseError, ResolveError

_OBJ = 'OBJ'
_TEST = 'TEST'
_LIST_TEST = 'LIST_TEST'
_LOGICAL = 'LOGICAL'
_OPTION = 'OPTION'

_VALUE = 'VALUE'
_IF = 'IF'
_AND = 'AND'
_OR = 'OR'

_EQUAL = '=='
_NOT_EQUAL = '!='

_IGNORE_ERRORS = '+IgnoreErrors'
_ALL_ENVS = '+AllEnvs'

class Element(object):

    def __init__(self, expression, delimiter):
        self._delimiter = delimiter
        self._export_path = None
        self._parameter_path = None
        self._parameter_value = None
        self._export_path, self._parameter_path, self._parameter_value = self._get_vars(expression[0][1], self._export_path, self._parameter_path, self._parameter_value)
        self._export_path, self._parameter_path, self._parameter_value = self._get_vars(expression[2][1], self._export_path, self._parameter_path, self._parameter_value)

        try:
            self._export_path.drop_first()
        except AttributeError:
            raise ExpressionError('No export')

        self._inv_refs = [ self._export_path ]
        self._test = expression[1][1]

        if self._parameter_path is not None:
            self._parameter_path.drop_first()
            self._refs = [ str(self._parameter_path) ]
        else:
            self._refs = []

    def refs(self):
        return self._refs

    def inv_refs(self):
        return self._inv_refs

    def value(self, context, items):
        if self._parameter_path is not None:
            self._parameter_value = self._resolve(self._parameter_path, context)

        if self._parameter_value is None or self._test is None:
            raise ExpressionError('Failed to render %s' % str(self), tbFlag=False)

        if self._export_path.exists_in(items):
            result = False
            export_value = self._resolve(self._export_path, items)
            if self._test == _EQUAL:
                if export_value == self._parameter_value:
                    result = True
            elif self._test == _NOT_EQUAL:
                if export_value != self._parameter_value:
                    result = True
            else:
                raise ExpressionError('Unknown test {0}'.format(self._test), tbFlag=False)
            return result
        else:
            return False

    def _resolve(self, path, dictionary):
        try:
            return path.get_value(dictionary)
        except KeyError as e:
            raise ResolveError(str(path))

    def _get_vars(self, var, export, parameter, value):
        if isinstance(var, str):
            path = DictPath(self._delimiter, var)
            if path.path[0].lower() == 'exports':
                export = path
            elif path.path[0].lower() == 'self':
                parameter = path
            elif path.path[0].lower() == 'true':
                value = True
            elif path.path[0].lower() == 'false':
                value = False
            else:
                value = var
        else:
            value = var
        return export, parameter, value


class Question(object):

    def __init__(self, expression, delimiter):
        self._elements = []
        self._operators = []
        self._delimiter = delimiter
        self._refs = []
        self._inv_refs = []
        i = 0
        while i < len(expression):
            e = Element(expression[i:], self._delimiter)
            self._elements.append(e)
            self._refs.extend(e.refs())
            self._inv_refs.extend(e.inv_refs())
            i += 3
            if i < len(expression):
                self._operators.append(expression[i][1])
                i += 1

    def refs(self):
        return self._refs

    def inv_refs(self):
        return self._inv_refs

    def value(self, context, items):
        if len(self._elements) == 0:
            return True
        elif len(self._elements) == 1:
            return self._elements[0].value(context, items)
        else:
            result = self._elements[0].value(context, items)
            for i in range(0, len(self._elements)-1):
                next_result = self._elements[i+1].value(context, items)
                if self._operators[i] == _AND:
                    result = result and next_result
                elif self._operators[i] == _OR:
                    result = result or next_result
                else:
                    raise ExpressionError('Unknown operator {0} {1}'.format(self._operators[i], self.elements), tbFlag=False)
            return result


class InvItem(Item):

    def _get_parser():

        def _object(string, location, tokens):
            token = tokens[0]
            tokens[0] = (_OBJ, token)

        def _integer(string, location, tokens):
            try:
                token = int(tokens[0])
            except ValueError:
                token = tokens[0]
            tokens[0] = (_OBJ, token)

        def _number(string, location, tokens):
            try:
                token = float(tokens[0])
            except ValueError:
                token = tokens[0]
            tokens[0] = (_OBJ, token)

        def _option(string, location, tokens):
            token = tokens[0]
            tokens[0] = (_OPTION, token)

        def _test(string, location, tokens):
            token = tokens[0]
            tokens[0] = (_TEST, token)

        def _logical(string, location, tokens):
            token = tokens[0]
            tokens[0] = (_LOGICAL, token)

        def _if(string, location, tokens):
            token = tokens[0]
            tokens[0] = (_IF, token)

        def _expr_var(string, location, tokens):
            token = tokens[0]
            tokens[0] = (_VALUE, token)

        def _expr_test(string, location, tokens):
            token = tokens[0]
            tokens[0] = (_TEST, token)

        def _expr_list_test(string, location, tokens):
            token = tokens[0]
            tokens[0] = (_LIST_TEST, token)

        white_space = pp.White().suppress()
        end = pp.StringEnd()
        ignore_errors = pp.CaselessLiteral(_IGNORE_ERRORS)
        all_envs = pp.CaselessLiteral(_ALL_ENVS)
        option = (ignore_errors | all_envs).setParseAction(_option)
        options = pp.Group(pp.ZeroOrMore(option + white_space))
        operator_test = (pp.Literal(_EQUAL) | pp.Literal(_NOT_EQUAL)).setParseAction(_test)
        operator_logical = (pp.CaselessLiteral(_AND) | pp.CaselessLiteral(_OR)).setParseAction(_logical)
        begin_if = pp.CaselessLiteral(_IF, ).setParseAction(_if)
        obj = pp.Word(pp.printables).setParseAction(_object)
        integer = pp.Word('0123456789-').setParseAction(_integer)
        number = pp.Word('0123456789-.').setParseAction(_number)
        item = integer | number | obj
        single_test = white_space + item + white_space + operator_test + white_space + item
        additional_test = white_space + operator_logical + single_test
        expr_var = pp.Group(obj + pp.Optional(white_space) + end).setParseAction(_expr_var)
        expr_test = pp.Group(obj + white_space + begin_if + single_test + pp.ZeroOrMore(additional_test) + end).setParseAction(_expr_test)
        expr_list_test = pp.Group(begin_if + single_test + pp.ZeroOrMore(additional_test) + end).setParseAction(_expr_list_test)
        expr = (expr_test | expr_var | expr_list_test)
        line = options + expr + end
        return line

    _parser = _get_parser()

    def __init__(self, item, settings):
        self.type = Item.INV_QUERY
        self._settings = settings
        self._needs_all_envs = False
        self._ignore_failed_render = self._settings.inventory_ignore_failed_render
        self._expr_text = item.render(None, None)
        self._parse_expression(self._expr_text)

    def _parse_expression(self, expr):
        try:
            tokens = InvItem._parser.parseString(expr).asList()
        except pp.ParseException as e:
            raise ParseError(e.msg, e.line, e.col, e.lineno)

        if len(tokens) == 1:
            self._expr_type = tokens[0][0]
            self._expr = list(tokens[0][1])
        elif len(tokens) == 2:
            for opt in tokens[0]:
                if opt[1] == _IGNORE_ERRORS:
                    self._ignore_failed_render = True
                elif opt[1] == _ALL_ENVS:
                    self._needs_all_envs = True
            self._expr_type = tokens[1][0]
            self._expr = list(tokens[1][1])
        else:
            raise ExpressionError('Failed to parse %s' % str(tokens), tbFlag=False)

        if self._expr_type == _VALUE:
            self._value_path = DictPath(self._settings.delimiter, self._expr[0][1]).drop_first()
            self._question = Question([], self._settings.delimiter)
            self._refs = []
            self._inv_refs = [ self._value_path ]
        elif self._expr_type == _TEST:
            self._value_path = DictPath(self._settings.delimiter, self._expr[0][1]).drop_first()
            self._question = Question(self._expr[2:], self._settings.delimiter)
            self._refs = self._question.refs()
            self._inv_refs = self._question.inv_refs()
            self._inv_refs.append(self._value_path)
        elif self._expr_type == _LIST_TEST:
            self._value_path = None
            self._question = Question(self._expr[1:], self._settings.delimiter)
            self._refs = self._question.refs()
            self._inv_refs = self._question.inv_refs()
        else:
            raise ExpressionError('Unknown expression type: %s' % self._expr_type, tbFlag=False)

    def assembleRefs(self, context):
        return

    def contents(self):
        return self._expr_text

    def has_inv_query(self):
        return True

    def has_references(self):
        return len(self._question.refs()) > 0

    def get_references(self):
        return self._question.refs()

    def get_inv_references(self):
        return self._inv_refs

    def needs_all_envs(self):
        return self._needs_all_envs

    def ignore_failed_render(self):
        return self._ignore_failed_render

    def _resolve(self, path, dictionary):
        try:
            return path.get_value(dictionary)
        except KeyError as e:
            raise ResolveError(str(path))

    def _value_expression(self, inventory):
        results = {}
        for (node, items) in iteritems(inventory):
            if self._value_path.exists_in(items):
                results[node] = copy.deepcopy(self._resolve(self._value_path, items))
        return results

    def _test_expression(self, context, inventory):
        if self._value_path is None:
            ExpressionError('Failed to render %s' % str(self), tbFlag=False)

        results = {}
        for (node, items) in iteritems(inventory):
            if self._question.value(context, items) and self._value_path.exists_in(items):
                results[node] = copy.deepcopy(self._resolve(self._value_path, items))
        return results

    def _list_test_expression(self, context, inventory):
        results = []
        for (node, items) in iteritems(inventory):
            if self._question.value(context, items):
                results.append(node)
        return results

    def render(self, context, inventory):
        if self._expr_type == _VALUE:
            return self._value_expression(inventory)
        elif self._expr_type == _TEST:
            return self._test_expression(context, inventory)
        elif self._expr_type == _LIST_TEST:
            return self._list_test_expression(context, inventory)
        raise ExpressionError('Failed to render %s' % str(self), tbFlag=False)

    def __str__(self):
        return ' '.join(str(j) for i,j in self._expr)

    def __repr__(self):
        return 'InvItem(%r)' % self._expr



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#

import copy

from six import iteritems

from reclass.settings import Settings
from reclass.datatypes import Parameters
from reclass.errors import InfiniteRecursionError, InterpolationError, ResolveError, ResolveErrorList
import unittest
try:
    import unittest.mock as mock
except ImportError:
    import mock

SIMPLE = {'one': 1, 'two': 2, 'three': 3}
SETTINGS = Settings()

class MockDevice(object):
    def __init__(self):
        self._text = ''

    def write(self, s):
        self._text += s
        return

    def text(self):
        return self._text

class TestParameters(unittest.TestCase):

    def _construct_mocked_params(self, iterable=None, settings=SETTINGS):
        p = Parameters(iterable, settings, '')
        self._base = base = p._base
        p._base = mock.MagicMock(spec_set=dict, wraps=base)
        p._base.__repr__ = mock.MagicMock(autospec=dict.__repr__,
                                          return_value=repr(base))
        p._base.__getitem__.side_effect = base.__getitem__
        p._base.__setitem__.side_effect = base.__setitem__
        return p, p._base

    def test_len_empty(self):
        p, b = self._construct_mocked_params()
        l = 0
        b.__len__.return_value = l
        self.assertEqual(len(p), l)
        b.__len__.assert_called_with()

    def test_constructor(self):
        p, b = self._construct_mocked_params(SIMPLE)
        l = len(SIMPLE)
        b.__len__.return_value = l
        self.assertEqual(len(p), l)
        b.__len__.assert_called_with()

    def test_repr_empty(self):
        p, b = self._construct_mocked_params()
        b.__repr__.return_value = repr({})
        self.assertEqual('%r' % p, '%s(%r)' % (p.__class__.__name__, {}))
        b.__repr__.assert_called_once_with()

    def test_repr(self):
        p, b = self._construct_mocked_params(SIMPLE)
        b.__repr__.return_value = repr(SIMPLE)
        self.assertEqual('%r' % p, '%s(%r)' % (p.__class__.__name__, SIMPLE))
        b.__repr__.assert_called_once_with()

    def test_equal_empty(self):
        p1, b1 = self._construct_mocked_params()
        p2, b2 = self._construct_mocked_params()
        b1.__eq__.return_value = True
        self.assertEqual(p1, p2)
        b1.__eq__.assert_called_once_with(b2)

    def test_equal_default_delimiter(self):
        p1, b1 = self._construct_mocked_params(SIMPLE)
        p2, b2 = self._construct_mocked_params(SIMPLE, SETTINGS)
        b1.__eq__.return_value = True
        self.assertEqual(p1, p2)
        b1.__eq__.assert_called_once_with(b2)

    def test_equal_contents(self):
        p1, b1 = self._construct_mocked_params(SIMPLE)
        p2, b2 = self._construct_mocked_params(SIMPLE)
        b1.__eq__.return_value = True
        self.assertEqual(p1, p2)
        b1.__eq__.assert_called_once_with(b2)

    def test_unequal_content(self):
        p1, b1 = self._construct_mocked_params()
        p2, b2 = self._construct_mocked_params(SIMPLE)
        b1.__eq__.return_value = False
        self.assertNotEqual(p1, p2)
        b1.__eq__.assert_called_once_with(b2)

    def test_unequal_delimiter(self):
        settings1 = Settings({'delimiter': ':'})
        settings2 = Settings({'delimiter': '%'})
        p1, b1 = self._construct_mocked_params(settings=settings1)
        p2, b2 = self._construct_mocked_params(settings=settings2)
        b1.__eq__.return_value = False
        self.assertNotEqual(p1, p2)
        b1.__eq__.assert_called_once_with(b2)

    def test_unequal_types(self):
        p1, b1 = self._construct_mocked_params()
        self.assertNotEqual(p1, None)
        self.assertEqual(b1.__eq__.call_count, 0)

    def test_construct_wrong_type(self):
        with self.assertRaises(TypeError):
            self._construct_mocked_params('wrong type')

    def test_merge_wrong_type(self):
        p, b = self._construct_mocked_params()
        with self.assertRaises(TypeError):
            p.merge('wrong type')

    def test_get_dict(self):
        p, b = self._construct_mocked_params(SIMPLE)
        p.initialise_interpolation()
        self.assertDictEqual(p.as_dict(), SIMPLE)

    def test_merge_scalars(self):
        p1, b1 = self._construct_mocked_params(SIMPLE)
        mergee = {'five':5,'four':4,'None':None,'tuple':(1,2,3)}
        p2, b2 = self._construct_mocked_params(mergee)
        p1.merge(p2)
        p1.initialise_interpolation()
        for (key, value) in iteritems(mergee):
            # check that each key, value in mergee resulted in a get call and
            # a __setitem__ call against b1 (the merge target)
            self.assertIn(mock.call(key), b1.get.call_args_list)
            self.assertIn(mock.call(key, value), b1.__setitem__.call_args_list)

    def test_stray_occurrence_overwrites_during_interpolation(self):
        p1 = Parameters({'r' : mock.sentinel.ref, 'b': '${r}'}, SETTINGS, '')
        p2 = Parameters({'b' : mock.sentinel.goal}, SETTINGS, '')
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict()['b'], mock.sentinel.goal)


class TestParametersNoMock(unittest.TestCase):

    def test_merge_scalars(self):
        p = Parameters(SIMPLE, SETTINGS, '')
        mergee = {'five':5,'four':4,'None':None,'tuple':(1,2,3)}
        p.merge(mergee)
        p.initialise_interpolation()
        goal = SIMPLE.copy()
        goal.update(mergee)
        self.assertDictEqual(p.as_dict(), goal)

    def test_merge_scalars_overwrite(self):
        p = Parameters(SIMPLE, SETTINGS, '')
        mergee = {'two':5,'four':4,'three':None,'one':(1,2,3)}
        p.merge(mergee)
        p.initialise_interpolation()
        goal = SIMPLE.copy()
        goal.update(mergee)
        self.assertDictEqual(p.as_dict(), goal)

    def test_merge_lists(self):
        l1 = [1,2,3]
        l2 = [2,3,4]
        p1 = Parameters(dict(list=l1[:]), SETTINGS, '')
        p2 = Parameters(dict(list=l2), SETTINGS, '')
        p1.merge(p2)
        p1.initialise_interpolation()
        self.assertListEqual(p1.as_dict()['list'], l1+l2)

    def test_merge_list_into_scalar(self):
        settings = Settings({'allow_list_over_scalar': True})
        l = ['foo', 1, 2]
        p1 = Parameters(dict(key=l[0]), settings, '')
        p2 = Parameters(dict(key=l[1:]), settings, '')
        p1.merge(p2)
        p1.initialise_interpolation()
        self.assertListEqual(p1.as_dict()['key'], l)

    def test_merge_scalar_over_list(self):
        l = ['foo', 1, 2]
        settings = Settings({'allow_scalar_over_list': True})
        p1 = Parameters(dict(key=l[:2]), settings, '')
        p2 = Parameters(dict(key=l[2]), settings, '')
        p1.merge(p2)
        p1.initialise_interpolation()
        self.assertEqual(p1.as_dict()['key'], l[2])

    def test_merge_none_over_list(self):
        l = ['foo', 1, 2]
        settings = Settings({'allow_none_override': True})
        p1 = Parameters(dict(key=l[:2]), settings, '')
        p2 = Parameters(dict(key=None), settings, '')
        p1.merge(p2)
        p1.initialise_interpolation()
        self.assertEqual(p1.as_dict()['key'], None)

    def test_merge_none_over_list_negative(self):
        l = ['foo', 1, 2]
        settings = Settings({'allow_none_override': False})
        p1 = Parameters(dict(key=l[:2]), settings, '')
        p2 = Parameters(dict(key=None), settings, '')
        with self.assertRaises(TypeError):
            p1.merge(p2)
            p1.initialise_interpolation()

    def test_merge_none_over_dict(self):
        settings = Settings({'allow_none_override': True})
        p1 = Parameters(dict(key=SIMPLE), settings, '')
        p2 = Parameters(dict(key=None), settings, '')
        p1.merge(p2)
        p1.initialise_interpolation()
        self.assertEqual(p1.as_dict()['key'], None)

    def test_merge_none_over_dict_negative(self):
        settings = Settings({'allow_none_override': False})
        p1 = Parameters(dict(key=SIMPLE), settings, '')
        p2 = Parameters(dict(key=None), settings, '')
        with self.assertRaises(TypeError):
            p1.merge(p2)
            p1.initialise_interpolation()

    # def test_merge_bare_dict_over_dict(self):
        # settings = Settings({'allow_bare_override': True})
        # p1 = Parameters(dict(key=SIMPLE), settings, '')
        # p2 = Parameters(dict(key=dict()), settings, '')
        # p1.merge(p2)
        # p1.initialise_interpolation()
        # self.assertEqual(p1.as_dict()['key'], {})

    # def test_merge_bare_list_over_list(self):
        # l = ['foo', 1, 2]
        # settings = Settings({'allow_bare_override': True})
        # p1 = Parameters(dict(key=l), settings, '')
        # p2 = Parameters(dict(key=list()), settings, '')
        # p1.merge(p2)
        # p1.initialise_interpolation()
        # self.assertEqual(p1.as_dict()['key'], [])

    def test_merge_dicts(self):
        mergee = {'five':5,'four':4,'None':None,'tuple':(1,2,3)}
        p = Parameters(dict(dict=SIMPLE), SETTINGS, '')
        p2 = Parameters(dict(dict=mergee), SETTINGS, '')
        p.merge(p2)
        p.initialise_interpolation()
        goal = SIMPLE.copy()
        goal.update(mergee)
        self.assertDictEqual(p.as_dict(), dict(dict=goal))

    def test_merge_dicts_overwrite(self):
        mergee = {'two':5,'four':4,'three':None,'one':(1,2,3)}
        p = Parameters(dict(dict=SIMPLE), SETTINGS, '')
        p2 = Parameters(dict(dict=mergee), SETTINGS, '')
        p.merge(p2)
        p.initialise_interpolation()
        goal = SIMPLE.copy()
        goal.update(mergee)
        self.assertDictEqual(p.as_dict(), dict(dict=goal))

    def test_merge_dicts_override(self):
        """Validate that tilde merge overrides function properly."""
        mergee = {'~one': {'a': 'alpha'},
                  '~two': ['gamma']}
        base = {'one': {'b': 'beta'},
                'two': ['delta']}
        goal = {'one': {'a': 'alpha'},
                'two': ['gamma']}
        p = Parameters(dict(dict=base), SETTINGS, '')
        p2 = Parameters(dict(dict=mergee), SETTINGS, '')
        p.merge(p2)
        p.initialise_interpolation()
        self.assertDictEqual(p.as_dict(), dict(dict=goal))

    def test_merge_dict_into_scalar(self):
        p = Parameters(dict(base='foo'), SETTINGS, '')
        p2 = Parameters(dict(base=SIMPLE), SETTINGS, '')
        with self.assertRaises(TypeError):
            p.merge(p2)
            p.interpolate()

    def test_merge_scalar_over_dict(self):
        settings = Settings({'allow_scalar_over_dict': True})
        p = Parameters(dict(base=SIMPLE), settings, '')
        mergee = {'base':'foo'}
        p2 = Parameters(mergee, settings, '')
        p.merge(p2)
        p.initialise_interpolation()
        self.assertDictEqual(p.as_dict(), mergee)

    def test_interpolate_single(self):
        v = 42
        d = {'foo': 'bar'.join(SETTINGS.reference_sentinels),
             'bar': v}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_interpolate_multiple(self):
        v = '42'
        d = {'foo': 'bar'.join(SETTINGS.reference_sentinels) + 'meep'.join(SETTINGS.reference_sentinels),
             'bar': v[0],
             'meep': v[1]}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_interpolate_multilevel(self):
        v = 42
        d = {'foo': 'bar'.join(SETTINGS.reference_sentinels),
             'bar': 'meep'.join(SETTINGS.reference_sentinels),
             'meep': v}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_interpolate_list(self):
        l = [41, 42, 43]
        d = {'foo': 'bar'.join(SETTINGS.reference_sentinels),
             'bar': l}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict()['foo'], l)

    def test_interpolate_infrecursion(self):
        v = 42
        d = {'foo': 'bar'.join(SETTINGS.reference_sentinels),
             'bar': 'foo'.join(SETTINGS.reference_sentinels)}
        p = Parameters(d, SETTINGS, '')
        with self.assertRaises(InfiniteRecursionError):
            p.interpolate()

    def test_nested_references(self):
        d = {'a': '${${z}}', 'b': 2, 'z': 'b'}
        r = {'a': 2, 'b': 2, 'z': 'b'}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict(), r)

    def test_nested_deep_references(self):
        d = {'one': { 'a': 1, 'b': '${one:${one:c}}', 'c': 'a' } }
        r = {'one': { 'a': 1, 'b': 1, 'c': 'a'} }
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict(), r)

    def test_stray_occurrence_overwrites_during_interpolation(self):
        p1 = Parameters({'r' : 1, 'b': '${r}'}, SETTINGS, '')
        p2 = Parameters({'b' : 2}, SETTINGS, '')
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict()['b'], 2)

    def test_referenced_dict_deep_overwrite(self):
        p1 = Parameters({'alpha': {'one': {'a': 1, 'b': 2} } }, SETTINGS, '')
        p2 = Parameters({'beta': '${alpha}'}, SETTINGS, '')
        p3 = Parameters({'alpha': {'one': {'c': 3, 'd': 4} },
                         'beta':  {'one': {'a': 99} } }, SETTINGS, '')
        r = {'alpha': {'one': {'a':1, 'b': 2, 'c': 3, 'd':4} },
             'beta': {'one': {'a':99, 'b': 2, 'c': 3, 'd':4} } }
        p1.merge(p2)
        p1.merge(p3)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_complex_reference_overwriting(self):
        p1 = Parameters({'one': 'abc_123_${two}_${three}', 'two': 'XYZ', 'four': 4}, SETTINGS, '')
        p2 = Parameters({'one': 'QWERTY_${three}_${four}', 'three': '999'}, SETTINGS, '')
        r = {'one': 'QWERTY_999_4', 'two': 'XYZ', 'three': '999', 'four': 4}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_nested_reference_with_overwriting(self):
        p1 = Parameters({'one': {'a': 1, 'b': 2, 'z': 'a'},
                         'two': '${one:${one:z}}' }, SETTINGS, '')
        p2 = Parameters({'one': {'z': 'b'} }, SETTINGS, '')
        r = {'one': {'a': 1, 'b':2, 'z': 'b'}, 'two': 2}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_merge_referenced_lists(self):
        p1 = Parameters({'one': [ 1, 2, 3 ], 'two': [ 4, 5, 6 ], 'three': '${one}'}, SETTINGS, '')
        p2 = Parameters({'three': '${two}'}, SETTINGS, '')
        r = {'one': [ 1, 2, 3 ], 'two': [ 4, 5, 6], 'three': [ 1, 2, 3, 4, 5, 6 ]}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_merge_referenced_dicts(self):
        p1 = Parameters({'one': {'a': 1, 'b': 2}, 'two': {'c': 3, 'd': 4}, 'three': '${one}'}, SETTINGS, '')
        p2 = Parameters({'three': '${two}'}, SETTINGS, '')
        r = {'one': {'a': 1, 'b': 2}, 'two': {'c': 3, 'd': 4}, 'three': {'a': 1, 'b': 2, 'c': 3, 'd': 4}}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_deep_refs_in_referenced_dicts(self):
        p = Parameters({'A': '${C:a}', 'B': {'a': 1, 'b': 2}, 'C': '${B}'}, SETTINGS, '')
        r = {'A': 1, 'B': {'a': 1, 'b': 2}, 'C': {'a': 1, 'b': 2}}
        p.interpolate()
        self.assertEqual(p.as_dict(), r)

    def test_overwrite_none(self):
        p1 = Parameters({'A': None, 'B': None, 'C': None, 'D': None, 'E': None, 'F': None}, SETTINGS, '')
        p2 = Parameters({'A': 'abc', 'B': [1, 2, 3], 'C': {'a': 'aaa', 'b': 'bbb'}, 'D': '${A}', 'E': '${B}', 'F': '${C}'}, SETTINGS, '')
        r = {'A': 'abc', 'B': [1, 2, 3], 'C': {'a': 'aaa', 'b': 'bbb'}, 'D': 'abc', 'E': [1, 2, 3], 'F': {'a': 'aaa', 'b': 'bbb'}}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_interpolate_escaping(self):
        v = 'bar'.join(SETTINGS.reference_sentinels)
        d = {'foo': SETTINGS.escape_character + 'bar'.join(SETTINGS.reference_sentinels),
             'bar': 'unused'}
        p = Parameters(d, SETTINGS, '')
        p.initialise_interpolation()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_interpolate_double_escaping(self):
        v = SETTINGS.escape_character + 'meep'
        d = {'foo': SETTINGS.escape_character + SETTINGS.escape_character + 'bar'.join(SETTINGS.reference_sentinels),
             'bar': 'meep'}
        p = Parameters(d, SETTINGS, '')
        p.interpolate()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_interpolate_escaping_backwards_compatibility(self):
        """In all following cases, escaping should not happen and the escape character
        needs to be printed as-is, to ensure backwards compatibility to older versions."""
        v = ' '.join([
            # Escape character followed by unescapable character
            '1', SETTINGS.escape_character,
            # Escape character followed by escape character
            '2', SETTINGS.escape_character + SETTINGS.escape_character,
            # Escape character followed by interpolation end sentinel
            '3', SETTINGS.escape_character + SETTINGS.reference_sentinels[1],
            # Escape character at the end of the string
            '4', SETTINGS.escape_character
            ])
        d = {'foo': v}
        p = Parameters(d, SETTINGS, '')
        p.initialise_interpolation()
        self.assertEqual(p.as_dict()['foo'], v)

    def test_escape_close_in_ref(self):
        p1 = Parameters({'one}': 1, 'two': '${one\\}}'}, SETTINGS, '')
        r = {'one}': 1, 'two': 1}
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_double_escape_in_ref(self):
        d = {'one\\': 1, 'two': '${one\\\\}'}
        p1 = Parameters(d, SETTINGS, '')
        r = {'one\\': 1, 'two': 1}
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_merging_for_multiple_nodes(self):
        p1 = Parameters({ 'alpha': { 'one': 111 }}, SETTINGS, '')
        p2 = Parameters({ 'beta': {'two': '${alpha:one}' }}, SETTINGS, '')
        p3 = Parameters({ 'beta': {'two': 222 }}, SETTINGS, '')
        n1 = Parameters({ 'name': 'node1'}, SETTINGS, '')
        r1 = { 'alpha': { 'one': 111 }, 'beta': { 'two': 111 }, 'name': 'node1' }
        r2 = { 'alpha': { 'one': 111 }, 'beta': { 'two': 222 }, 'name': 'node2' }
        n1.merge(p1)
        n1.merge(p2)
        n1.interpolate()
        n2 = Parameters({'name': 'node2'}, SETTINGS, '')
        n2.merge(p1)
        n2.merge(p2)
        n2.merge(p3)
        n2.interpolate()
        self.assertEqual(n1.as_dict(), r1)
        self.assertEqual(n2.as_dict(), r2)

    def test_list_merging_for_multiple_nodes(self):
        p1 = Parameters({ 'alpha': { 'one': [1, 2] }}, SETTINGS, '')
        p2 = Parameters({ 'beta': {'two': '${alpha:one}' }}, SETTINGS, '')
        p3 = Parameters({ 'beta': {'two': [3] }}, SETTINGS, '')
        n1 = Parameters({ 'name': 'node1'}, SETTINGS, '')
        r1 = { 'alpha': { 'one': [1, 2] }, 'beta': { 'two': [1, 2] }, 'name': 'node1' }
        r2 = { 'alpha': { 'one': [1, 2] }, 'beta': { 'two': [1, 2, 3] }, 'name': 'node2' }
        n1.merge(p1)
        n1.merge(p2)
        n1.interpolate()
        n2 = Parameters({'name': 'node2'}, SETTINGS, '')
        n2.merge(p1)
        n2.merge(p2)
        n2.merge(p3)
        n2.interpolate()
        self.assertEqual(n1.as_dict(), r1)
        self.assertEqual(n2.as_dict(), r2)

    def test_dict_merging_for_multiple_nodes(self):
        p1 = Parameters({ 'alpha': { 'one': { 'a': 'aa', 'b': 'bb' }}}, SETTINGS, '')
        p2 = Parameters({ 'beta': {'two': '${alpha:one}' }}, SETTINGS, '')
        p3 = Parameters({ 'beta': {'two': {'c': 'cc' }}}, SETTINGS, '')
        n1 = Parameters({ 'name': 'node1'}, SETTINGS, '')
        r1 = { 'alpha': { 'one': {'a': 'aa', 'b': 'bb'} }, 'beta': { 'two': {'a': 'aa', 'b': 'bb'} }, 'name': 'node1' }
        r2 = { 'alpha': { 'one': {'a': 'aa', 'b': 'bb'} }, 'beta': { 'two': {'a': 'aa', 'b': 'bb', 'c': 'cc'} }, 'name': 'node2' }
        n1.merge(p1)
        n1.merge(p2)
        n1.interpolate()
        n2 = Parameters({'name': 'node2'}, SETTINGS, '')
        n2.merge(p1)
        n2.merge(p2)
        n2.merge(p3)
        n2.interpolate()
        self.assertEqual(n1.as_dict(), r1)
        self.assertEqual(n2.as_dict(), r2)

    def test_list_merging_with_refs_for_multiple_nodes(self):
        p1 = Parameters({ 'alpha': { 'one': [1, 2], 'two': [3, 4] }}, SETTINGS, '')
        p2 = Parameters({ 'beta': { 'three': '${alpha:one}' }}, SETTINGS, '')
        p3 = Parameters({ 'beta': { 'three': '${alpha:two}' }}, SETTINGS, '')
        p4 = Parameters({ 'beta': { 'three': '${alpha:one}' }}, SETTINGS, '')
        n1 = Parameters({ 'name': 'node1' }, SETTINGS, '')
        r1 = {'alpha': {'one': [1, 2], 'two': [3, 4]}, 'beta': {'three': [1, 2]}, 'name': 'node1'}
        r2 = {'alpha': {'one': [1, 2], 'two': [3, 4]}, 'beta': {'three': [1, 2, 3, 4, 1, 2]}, 'name': 'node2'}
        n2 = Parameters({ 'name': 'node2' }, SETTINGS, '')
        n2.merge(p1)
        n2.merge(p2)
        n2.merge(p3)
        n2.merge(p4)
        n2.interpolate()
        n1.merge(p1)
        n1.merge(p2)
        n1.interpolate()
        self.assertEqual(n1.as_dict(), r1)
        self.assertEqual(n2.as_dict(), r2)

    def test_nested_refs_with_multiple_nodes(self):
        p1 = Parameters({ 'alpha': { 'one': 1, 'two': 2 } }, SETTINGS, '')
        p2 = Parameters({ 'beta': { 'three': 'one' } }, SETTINGS, '')
        p3 = Parameters({ 'beta': { 'three': 'two' } }, SETTINGS, '')
        p4 = Parameters({ 'beta': { 'four': '${alpha:${beta:three}}' } }, SETTINGS, '')
        n1 = Parameters({ 'name': 'node1' }, SETTINGS, '')
        r1 = {'alpha': {'one': 1, 'two': 2}, 'beta': {'three': 'one', 'four': 1}, 'name': 'node1'}
        r2 = {'alpha': {'one': 1, 'two': 2}, 'beta': {'three': 'two', 'four': 2}, 'name': 'node2'}
        n1.merge(p1)
        n1.merge(p4)
        n1.merge(p2)
        n1.interpolate()
        n2 = Parameters({ 'name': 'node2' }, SETTINGS, '')
        n2.merge(p1)
        n2.merge(p4)
        n2.merge(p3)
        n2.interpolate()
        self.assertEqual(n1.as_dict(), r1)
        self.assertEqual(n2.as_dict(), r2)

    def test_nested_refs_error_message(self):
        # beta is missing, oops
        p1 = Parameters({'alpha': {'one': 1, 'two': 2}, 'gamma': '${alpha:${beta}}'}, SETTINGS, '')
        with self.assertRaises(InterpolationError) as error:
            p1.interpolate()
        self.assertEqual(error.exception.message, "-> \n   Bad references, at gamma\n      ${beta}")

    def test_multiple_resolve_errors(self):
        p1 = Parameters({'alpha': '${gamma}', 'beta': '${gamma}'}, SETTINGS, '')
        with self.assertRaises(ResolveErrorList) as error:
            p1.interpolate()
        self.assertEqual(error.exception.message, "-> \n   Cannot resolve ${gamma}, at alpha\n   Cannot resolve ${gamma}, at beta")

    def test_force_single_resolve_error(self):
        settings = copy.deepcopy(SETTINGS)
        settings.group_errors = False
        p1 = Parameters({'alpha': '${gamma}', 'beta': '${gamma}'}, settings, '')
        with self.assertRaises(ResolveError) as error:
            p1.interpolate()
        self.assertEqual(error.exception.message, "-> \n   Cannot resolve ${gamma}, at alpha")

    def test_ignore_overwriten_missing_reference(self):
        settings = copy.deepcopy(SETTINGS)
        settings.ignore_overwritten_missing_references = True
        p1 = Parameters({'alpha': '${beta}'}, settings, '')
        p2 = Parameters({'alpha': '${gamma}'}, settings, '')
        p3 = Parameters({'gamma': 3}, settings, '')
        r1 = {'alpha': 3, 'gamma': 3}
        p1.merge(p2)
        p1.merge(p3)
        err1 = "[WARNING] Reference '${beta}' undefined\n"
        with mock.patch('sys.stderr', new=MockDevice()) as std_err:
            p1.interpolate()
        self.assertEqual(p1.as_dict(), r1)
        self.assertEqual(std_err.text(), err1)

    def test_ignore_overwriten_missing_reference_last_value(self):
        # an error should be raised if the last reference to be merged
        # is missing even if ignore_overwritten_missing_references is true
        settings = copy.deepcopy(SETTINGS)
        settings.ignore_overwritten_missing_references = True
        p1 = Parameters({'alpha': '${gamma}'}, settings, '')
        p2 = Parameters({'alpha': '${beta}'}, settings, '')
        p3 = Parameters({'gamma': 3}, settings, '')
        p1.merge(p2)
        p1.merge(p3)
        with self.assertRaises(InterpolationError) as error:
            p1.interpolate()
        self.assertEqual(error.exception.message, "-> \n   Cannot resolve ${beta}, at alpha")

    def test_ignore_overwriten_missing_reference_dict(self):
        # setting ignore_overwritten_missing_references to true should
        # not change the behaviour for dicts
        settings = copy.deepcopy(SETTINGS)
        settings.ignore_overwritten_missing_references = True
        p1 = Parameters({'alpha': '${beta}'}, settings, '')
        p2 = Parameters({'alpha': '${gamma}'}, settings, '')
        p3 = Parameters({'gamma': {'one': 1, 'two': 2}}, settings, '')
        err1 = "[WARNING] Reference '${beta}' undefined\n"
        p1.merge(p2)
        p1.merge(p3)
        with self.assertRaises(InterpolationError) as error, mock.patch('sys.stderr', new=MockDevice()) as std_err:
            p1.interpolate()
        self.assertEqual(error.exception.message, "-> \n   Cannot resolve ${beta}, at alpha")
        self.assertEqual(std_err.text(), err1)

    def test_escaped_string_in_ref_dict_1(self):
        # test with escaped string in first dict to be merged
        p1 = Parameters({'a': { 'one': '${a_ref}' }, 'b': { 'two': '\${not_a_ref}' }, 'c': '${b}', 'a_ref': 123}, SETTINGS, '')
        p2 = Parameters({'c': '${a}'}, SETTINGS, '')
        r = { 'a': { 'one': 123 }, 'b': { 'two': '${not_a_ref}' }, 'c': { 'one': 123, 'two': '${not_a_ref}' }, 'a_ref': 123}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_escaped_string_in_ref_dict_2(self):
        # test with escaped string in second dict to be merged
        p1 = Parameters({'a': { 'one': '${a_ref}' }, 'b': { 'two': '\${not_a_ref}' }, 'c': '${a}', 'a_ref': 123}, SETTINGS, '')
        p2 = Parameters({'c': '${b}'}, SETTINGS, '')
        r = { 'a': { 'one': 123 }, 'b': { 'two': '${not_a_ref}' }, 'c': { 'one': 123, 'two': '${not_a_ref}' }, 'a_ref': 123}
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_complex_overwrites_1(self):
        # find a better name for this test
        p1 = Parameters({ 'test': { 'dict': { 'a': '${values:one}', 'b': '${values:two}' } },
                          'values': { 'one': 1, 'two': 2, 'three': { 'x': 'X', 'y': 'Y' } } }, SETTINGS, '')
        p2 = Parameters({ 'test': { 'dict': { 'c': '${values:two}' } } }, SETTINGS, '')
        p3 = Parameters({ 'test': { 'dict': { '~b': '${values:three}' } } }, SETTINGS, '')
        r = {'test': {'dict': {'a': 1, 'b': {'x': 'X', 'y': 'Y'}, 'c': 2}}, 'values': {'one': 1, 'three': {'x': 'X', 'y': 'Y'}, 'two': 2} }
        p2.merge(p3)
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_escaped_string_overwrites(self):
        p1 = Parameters({ 'test': '\${not_a_ref}' }, SETTINGS, '')
        p2 = Parameters({ 'test': '\${also_not_a_ref}' }, SETTINGS, '')
        r = { 'test': '${also_not_a_ref}' }
        p1.merge(p2)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)

    def test_escaped_string_in_ref_dict_overwrite(self):
        p1 = Parameters({'a': { 'one': '\${not_a_ref}' }, 'b': { 'two': '\${also_not_a_ref}' }}, SETTINGS, '')
        p2 = Parameters({'c': '${a}'}, SETTINGS, '')
        p3 = Parameters({'c': '${b}'}, SETTINGS, '')
        p4 = Parameters({'c': { 'one': '\${again_not_a_ref}' } }, SETTINGS, '')
        r = {'a': {'one': '${not_a_ref}'}, 'b': {'two': '${also_not_a_ref}'}, 'c': {'one': '${again_not_a_ref}', 'two': '${also_not_a_ref}'}}
        p1.merge(p2)
        p1.merge(p3)
        p1.merge(p4)
        p1.interpolate()
        self.assertEqual(p1.as_dict(), r)


if __name__ == '__main__':
    unittest.main()
